-module(irc_user).

-behaviour(gen_server).

%% API
-export([start_link/1]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).
	 
%% internal functions
-export([handle_irc/2, send/3, send/4, welcome/3, msg/4, quit/4, call_channels/2]).

%% hostent record
-include_lib("kernel/include/inet.hrl").

-record(state, {nick, user, host, welcomed, socket, channels}).

%%====================================================================
%% API
%%====================================================================
%%--------------------------------------------------------------------
%% Function: start_link() -> {ok,Pid} | ignore | {error,Error}
%% Description: Starts the server
%%--------------------------------------------------------------------
start_link(Ref) ->
    gen_server:start_link({global, Ref}, ?MODULE, [], []).
    
%%====================================================================
%% gen_server callbacks
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init(Args) -> {ok, State} |
%%                         {ok, State, Timeout} |
%%                         ignore               |
%%                         {stop, Reason}
%% Description: Initiates the server
%%--------------------------------------------------------------------
init([]) ->
    io:format("new user~n"),
    {ok, #state{}}.

%%--------------------------------------------------------------------
%% Function: %% handle_call(Request, From, State) -> {reply, Reply, State} |
%%                                      {reply, Reply, State, Timeout} |
%%                                      {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, Reply, State} |
%%                                      {stop, Reason, State}
%% Description: Handling call messages
%%--------------------------------------------------------------------
handle_call(nickuser, _From, State) ->
    Nick = State#state.nick,
    User = State#state.user,
    {reply, {Nick,User}, State};

handle_call(Request, _From, State) ->
    io:format("Call ~p not understood~n", [Request]),
    {reply, fail, State}.

%%--------------------------------------------------------------------
%% Function: handle_cast(Msg, State) -> {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, State}
%% Description: Handling cast messages
%%--------------------------------------------------------------------
handle_cast({create, Socket}, State) ->
    {ok, {Adress, Port}} = inet:peername(Socket),
    {ok, #hostent{h_name = Host}} = inet:gethostbyaddr(Adress),
    {noreply, #state{nick = State#state.nick,
		     user = State#state.user,
		     host = Host,
		     welcomed = false,
		     socket = Socket,
		     channels = []}};

handle_cast({send, Action, Args}, State) ->
    irc_user:send(State#state.socket,Action, Args),
    {noreply, State};

handle_cast({sendexp, Exp, Action, Args}, State) ->
    irc_user:send(State#state.socket,Exp, Action, Args),
    {noreply, State};

handle_cast({join, Pid}, State) ->
    {noreply, #state{nick = State#state.nick,
                     user = State#state.user,
                     host = State#state.host,
		     welcomed = State#state.welcomed,
                     socket = State#state.socket,
                     channels = [Pid|State#state.channels]}};

handle_cast(please_stop, State) ->
    io:format("Stop request~n"),
    {stop, normal, State};

handle_cast(Msg, State) ->
    io:format("Cast ~p not understood~n", [Msg]),
    {noreply, State}.

%%--------------------------------------------------------------------
%% Function: handle_info(Info, State) -> {noreply, State} |
%%                                       {noreply, State, Timeout} |
%%                                       {stop, Reason, State}
%% Description: Handling all non call/cast messages
%%--------------------------------------------------------------------
handle_info({tcp, Socket, Data}, State) ->
    io:format("tcp message : ~p~n",[Data]),
    irc_user:handle_irc(irc:parse_cmd(Data),State);

handle_info(Info, State) ->
    io:format("Info ~p not understood~n", [Info]),
    {noreply, State}.

%%--------------------------------------------------------------------
%% Function: terminate(Reason, State) -> void()
%% Description: This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any necessary
%% cleaning up. When it returns, the gen_server terminates with Reason.
%% The return value is ignored.
%%--------------------------------------------------------------------
terminate(_Reason, _State) ->
    io:format("user closed : ~p~n",[_Reason]),
    ok.

%%--------------------------------------------------------------------
%% Func: code_change(OldVsn, State, Extra) -> {ok, NewState}
%% Description: Convert process state when code is changed
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------
handle_irc({Sender, Action, Args},State) ->
    io:format("sender: ~p, action: ~p, args: ~p~n",[Sender, Action, Args]),
    case Action of
	"NICK" -> case gen_server:call({global,server},{nick,self(),hd(Args)}) of
		      ok -> {User, Welcome} = case State#state.user of
			     	                  undefined -> {undefined,
						                false};
		      	     		          Olduser -> gen_server:cast({global, server},{broadcast, Olduser, "NICK",Args}),
						             irc_user:call_channels(State#state.channels,
							                            {changenick,
								                     self(),
								                     hd(Args)}),	
				                             {hd(Args)++"!"++irc:user("nick!"++Olduser)++"@"++State#state.host,
						              true}
		  	                       end,
		  
		            Welcomed = case {State#state.welcomed, Welcome} of
		      	                   {false, true} -> irc_user:welcome(State#state.socket,
				  	                                     hd(Args),
					                                     User),
					                    true;
		                           _ -> State#state.welcomed
		                       end,
		            {noreply, #state{nick=hd(Args),
			                     user=User, 
			                     host=State#state.host,
			                     welcomed = Welcomed,
			                     socket=State#state.socket,
			                     channels=State#state.channels}};
		      fail -> irc_user:send(State#state.socket, "NOTICE",[hd(Args),"nick already used, choose an another one"]),
			      {noreply, State}
		  end;	
	"USER" -> Nick = State#state.nick,
		  Socket = State#state.socket,
		  {User, Welcome} = case Nick of
		                        undefined -> {"!"++hd(Args)++"@"++State#state.host,
						      false};
		                        _ -> {Nick++"!"++hd(Args)++"@"++State#state.host,
				              true}
		                    end,
                  Welcomed = case {State#state.welcomed, Welcome} of
		      	         {false, true} -> irc_user:welcome(Socket,Nick,User),
				                  true;
		                 _ -> State#state.welcomed
		  end,
		  {noreply, #state{nick=Nick,
				   user = User,
				   host = State#state.host,
				   welcomed = Welcomed,
				   socket = Socket,
				   channels = State#state.channels}};
	"JOIN" -> gen_server:cast({global,server},{channel, join, hd(Args), self()}),
		  {noreply, State};
        "PRIVMSG" -> irc_user:msg(State#state.user, hd(Args), Action, hd(tl(Args))),
		     {noreply, State};
	"TOPIC" -> case gen_server:call({global,server},{c_pid, hd(Args)}) of
		       {ok, Pid} -> gen_server:cast(Pid, {msg, State#state.user, Action, tl(Args)});
		       fail -> ok
                   end,
		   {noreply, State};
	"LIST" -> ok = gen_server:call({global, server},{list, self(), State#state.nick}),
		  {noreply, State};
   	"PART" -> Reason = case tl(Args) of
		      [R|_] -> R;
		      [] -> "no reason"
		  end,
  		  irc_user:msg(State#state.user, hd(Args), Action, Reason),
		  {ok, Pid} = gen_server:call({global, server},{c_pid, hd(Args)}),
		  {noreply, #state{nick = State#state.nick,
		 	           user = State#state.user,
			           host = State#state.host,
				   welcomed = State#state.welcomed,
			           socket = State#state.socket,
			           channels = lists:delete(Pid,State#state.channels)}};
	"QUIT" -> Reason = case Args of
		     	      [R|_] -> R;
			      [] -> "no reason"
			   end,
		  irc_user:quit(self(), State#state.user, State#state.channels, Reason),
	          {noreply, State};
	_ -> io:format("irc action not understood~n"),
	     {noreply, State}
    end;

handle_irc([Head|Tail],State) ->
    irc_user:handle_irc(Tail,irc_user:handle_irc(Head,State));
   
handle_irc([],State) ->
    State.

send(Socket, Action, Args) ->
    {ok, Server} = inet:gethostname(),
    irc_user:send(Socket, Server, Action, Args).

send(Socket, Exp, Action, Args) ->
    Data = ":"++Exp++" "++irc:format(Action,Args),
    io:format("sending on tcp: ~p~n",[Data]),
    gen_tcp:send(Socket, Data).

welcome(Socket, Nick, User) ->
    {ok, Server} = inet:gethostname(),
    irc_user:send(Socket, "001", [Nick, "Welcome to the Internet Relay Network, "++User]),
    irc_user:send(Socket, "002", [Nick, "your host is "++Server++", running version 1.0"]),
    irc_user:send(Socket, "003", [Nick, "this server wes created on Wed June 28 2007"]),
    irc_user:send(Socket, "375", [Nick, "- "++Server++" Message of the Day -"]),
    irc_user:send(Socket, "372", [Nick, "Nous n'avions pas les mêmes pensées mais nous avions des pensées de même couleur. (Jules Renard)"]),
    irc_user:send(Socket,"376", [Nick, "End of /MOTD command"]),
    irc_user:send(Socket,"NOTICE", [Nick, "type /join channel to join a channel"]),
    irc_user:send(Socket,"NOTICE", [Nick, "for example /join #fazol"]),
    irc_user:send(Socket,"NOTICE", [Nick, "this is a little server and only few commands are implemented"]),
    irc_user:send(Socket,"NOTICE", [Nick, "commands : /nick nickmane /user username /join #channel"]),
    irc_user:send(Socket,"NOTICE", [Nick, "commands : /list /quit reason"]).

msg(User, Dest, Action, Msg) ->
    case gen_server:call({global, server},{c_pid, Dest}) of
	{ok, Pid} -> gen_server:cast(Pid,{msg, User, Action, Msg});
        fail -> case gen_server:call({global, server},{u_pid, Dest}) of
	            {ok, Pid} -> gen_server:cast(Pid,{sendexp, User, Action, [Msg]});
	            fail -> io:format("not a chan and not a person~n")
	        end
    end.

quit(Pid, User,Channels, Msg) ->
    io:format("procedure pour quitter: ~p~n",[Channels]),
    case Channels of
	[C_pid|Tail] -> gen_server:cast(C_pid,{msg, User, "QUIT", Msg}),
		        irc_user:quit(Pid,User,Tail,Msg);
	[] -> ok
    end,
    gen_server:cast({global,server},{userexit,{Pid,irc:nick(User)}}),
    gen_server:cast(Pid, please_stop).

call_channels(Channels, Msg) ->
    case Channels of
	[Pid|Tail] -> gen_server:call(Pid, Msg),
		      irc_user:call_channels(Tail,Msg);
	[] -> ok
    end.
	
